# SPDX-FileCopyrightText: 2026 Greg Brandt <brandt.greg@gmail.com>
#
# SPDX-License-Identifier: Apache-2.0

from .browser import ChromeTool, ProxyTool
from .crawl import CrawlTool
from .parse import ParseTool, Parser

__all__ = [
    "ChromeTool",
    "CrawlTool",
    "ParseTool",
    "Parser",
    "ProxyTool",
]



# SPDX-FileCopyrightText: 2026 Greg Brandt <brandt.greg@gmail.com>
#
# SPDX-License-Identifier: Apache-2.0

from .base import Parser, list_parsers
from avrokit import URL, parse_url, create_url_mapping, avro_reader, avro_writer
from dataclasses import dataclass
import argparse
import importlib
import importlib.util
import logging
import sys
from pathlib import Path

logger = logging.getLogger("parsetool")


@dataclass
class ParseToolStats:
    count_input: int = 0
    count_output: int = 0
    count_error: int = 0


class ParseTool:
    def name(self) -> str:
        return "parse"

    def load_parser(self, class_name: str, script_path: str | None = None) -> Parser:
        if script_path:
            # Load module from arbitrary Python file
            script_file = Path(script_path).resolve()
            if not script_file.exists():
                raise FileNotFoundError(f"Script file not found: {script_file}")
            module_name = script_file.stem
            spec = importlib.util.spec_from_file_location(module_name, script_file)
            if spec is None or spec.loader is None:
                raise ImportError(f"Cannot load module from {script_file}")
            module = importlib.util.module_from_spec(spec)
            sys.modules[module_name] = module
            spec.loader.exec_module(module)
            return getattr(module, class_name)()
        else:
            # Load from installed package
            module_name, _, class_name = class_name.rpartition(".")
            module = importlib.import_module(module_name)
            return getattr(module, class_name)()

    def list_parsers(self) -> list[str]:
        acc: list[str] = []
        for parser in list_parsers():
            acc.append(".".join([parser.__module__, parser.__name__]))
        return acc

    def parse(
        self, parser: Parser, base_input_url: URL, base_output_url: URL
    ) -> ParseToolStats:
        # Stats accumulator
        stats = ParseToolStats()
        # Map the input and output URLs
        for input_url, output_url in create_url_mapping(
            base_input_url, base_output_url
        ):
            logger.info("Parsing %s -> %s", input_url, output_url)
            # Open the input URL and output URL as Avro files
            with (
                avro_reader(input_url.with_mode("rb")) as reader,
                avro_writer(output_url.with_mode("wb"), parser.schema()) as writer,
            ):
                for record in reader:
                    try:
                        stats.count_input += 1
                        for parsed_record in parser.parse(record):
                            if parsed_record is None:
                                logger.debug("Record is None, skipping")
                                continue
                            stats.count_output += 1
                            writer.append(parsed_record)
                    except Exception as e:
                        logger.error("Error parsing record: %s", e)
                        if logging.DEBUG == logger.getEffectiveLevel():
                            logger.exception(e)
                        stats.count_error += 1
                    finally:
                        if logging.DEBUG == logger.getEffectiveLevel():
                            if stats.count_input > 0 and stats.count_input % 100 == 0:
                                logger.debug("%s", stats)
        return stats

    def configure(self, subparsers: argparse._SubParsersAction) -> None:
        parser = subparsers.add_parser(self.name(), help="Parse output of crawl")
        parser.add_argument(
            "name",
            help="Name of the parser class to use (e.g., 'StandardPageParser' or 'rubbernecker.parse.standard.StandardPageParser')",
        )
        parser.add_argument(
            "input_url",
            help="URL containing result of crawl",
        )
        parser.add_argument(
            "output_url",
            help="URL to save the parsed data",
        )
        parser.add_argument(
            "--script",
            help="Path to a Python file containing the parser implementation",
        )

    def run(self, args: argparse.Namespace) -> None:
        # Validate built-in parser name if no script provided
        if not args.script and args.name not in self.list_parsers():
            raise ValueError(
                f"Unknown parser: {args.name}. "
                f"Use --script to load from a file, or choose from: {', '.join(self.list_parsers())}"
            )
        stats = self.parse(
            self.load_parser(args.name, args.script),
            parse_url(args.input_url),
            parse_url(args.output_url),
        )
        logger.info("%s (done)", stats)



# SPDX-FileCopyrightText: 2026 Greg Brandt <brandt.greg@gmail.com>
#
# SPDX-License-Identifier: Apache-2.0

import argparse
import asyncio
import base64
from dataclasses import dataclass
import logging
import contextlib
from typing import Tuple
from urllib.parse import urlparse

logger = logging.getLogger(__name__)


@dataclass
class ProxyAddress:
    host: str
    port: int
    auth: Tuple[str, str] | None = None

    def auth_token(self) -> str | None:
        if self.auth:
            username, password = self.auth
            return base64.b64encode(f"{username}:{password}".encode()).decode()
        return None


class ProxyTool:
    def name(self) -> str:
        return "proxy"

    def _create_connect_request(self, target: str, upstream_auth: str | None) -> str:
        if upstream_auth:
            return (
                f"CONNECT {target} HTTP/1.1\r\n"
                f"Host: {target}\r\n"
                f"Proxy-Connection: Keep-Alive\r\n"
                f"Proxy-Authorization: Basic {upstream_auth}\r\n"
                "\r\n"
            )
        else:
            return (
                f"CONNECT {target} HTTP/1.1\r\n"
                f"Host: {target}\r\n"
                f"Proxy-Connection: Keep-Alive\r\n"
                "\r\n"
            )

    def _create_handler(
        self,
        upstream: ProxyAddress,
    ):
        # Create HTTP basic authentication header for upstream proxy
        upstream_auth = upstream.auth_token()

        async def handle(reader: asyncio.StreamReader, writer: asyncio.StreamWriter):
            try:
                # Read the incoming request
                request_line = await reader.readuntil(b"\r\n")
                method, target, _ = request_line.decode().strip().split()
                headers = await reader.readuntil(b"\r\n\r\n")
                # Connect to upstream proxy
                try:
                    upstream_reader, upstream_writer = await asyncio.open_connection(
                        upstream.host, upstream.port
                    )
                except Exception as e:
                    logger.error("Upstream connection failed: %s", e)
                    writer.close()
                    await writer.wait_closed()
                    return
                if method.upper() == "CONNECT":
                    # Handle CONNECT method
                    connect_req = self._create_connect_request(target, upstream_auth)
                    upstream_writer.write(connect_req.encode())
                    await upstream_writer.drain()
                    # Relay response back to client
                    response = await upstream_reader.readuntil(b"\r\n\r\n")
                    writer.write(response)
                    await writer.drain()
                else:
                    # Proxy other HTTP methods
                    upstream_writer.write(
                        f"{method} {target} HTTP/1.1\r\n"
                        f"Proxy-Authorization: Basic {upstream_auth}\r\n".encode()
                        + headers
                    )
                    await upstream_writer.drain()

                async def pipe(src_reader, dst_writer):
                    try:
                        while True:
                            data = await src_reader.read(4096)
                            if not data:
                                break
                            dst_writer.write(data)
                            await dst_writer.drain()
                    except Exception as e:
                        logger.error("Pipe error: %s", e)
                    finally:
                        try:
                            dst_writer.close()
                            await dst_writer.wait_closed()
                        except Exception:
                            pass

                # Proxy streams
                task1 = asyncio.create_task(pipe(reader, upstream_writer))
                task2 = asyncio.create_task(pipe(upstream_reader, writer))
                done, pending = await asyncio.wait(
                    [task1, task2],
                    return_when=asyncio.FIRST_COMPLETED,
                )
                for task in done:
                    if task.exception():
                        logger.error("Task error: %s", task.exception())
                for task in pending:
                    task.cancel()
                    with contextlib.suppress(asyncio.CancelledError):
                        await task

            except Exception as e:
                logger.error("Handler error: %s", e)

        return handle

    async def start_proxy(self, upstream: ProxyAddress, listen: ProxyAddress):
        handler = self._create_handler(upstream)
        server = await asyncio.start_server(
            handler,
            listen.host,
            listen.port,
        )
        logger.info("Listening on %s:%d", listen.host, listen.port)
        async with server:
            await server.serve_forever()

    def configure(self, subparsers: argparse._SubParsersAction) -> None:
        parser = subparsers.add_parser(self.name(), help="Run an asyncio proxy server")
        parser.add_argument(
            "upstream",
            type=str,
            help="Upstream proxy URL",
        )
        parser.add_argument(
            "listen",
            type=str,
            help="Local proxy URL",
            nargs="?",
            default="127.0.0.1:3128",
        )

    def _parse_address(self, url: str) -> ProxyAddress:
        if "://" not in url:
            url = "//" + url  # Prepend dummy scheme if missing
        parsed = urlparse(url)
        if not parsed.hostname:
            raise ValueError("Invalid URL: missing hostname")
        if not parsed.port:
            raise ValueError("Invalid URL: missing port")
        if (
            parsed.username
            and not parsed.password
            or parsed.password
            and not parsed.username
        ):
            raise ValueError(
                "Invalid URL: must have both username and password for auth"
            )
        return ProxyAddress(
            host=parsed.hostname,
            port=parsed.port,
            auth=(
                (
                    parsed.username,
                    parsed.password,
                )
                if parsed.username and parsed.password
                else None
            ),
        )

    def run(self, args: argparse.Namespace) -> None:
        try:
            upstream = self._parse_address(args.upstream)
            listen = self._parse_address(args.listen)
            asyncio.run(self.start_proxy(upstream, listen))
        except KeyboardInterrupt:
            logger.info("Proxy shutdown requested by user.")



# SPDX-FileCopyrightText: 2026 Greg Brandt <brandt.greg@gmail.com>
#
# SPDX-License-Identifier: Apache-2.0

import hashlib
import math
from typing import Generator

# Tuned for false positive rate of 0.1% with 1 million elements
DEFAULT_SIZE = 15_000_000  # ~1.8 MB
DEFAULT_HASH_COUNT = 9


class BloomFilter:
    def __init__(
        self, size: int = DEFAULT_SIZE, hash_count: int = DEFAULT_HASH_COUNT
    ) -> None:
        """
        Initialize a Bloom filter with a given size and number of hash functions.

        :param size: Size of the bit array.
        :param hash_count: Number of hash functions to use.
        """
        self.size: int = size
        self.hash_count: int = hash_count
        self.bit_array: list[int] = [0] * size

    def __str__(self) -> str:
        return f"BloomFilter(size={self.size}, hash_count={self.hash_count})"

    def _hashes(self, item: str) -> Generator[int, None, None]:
        """
        Generate hash values for the given item using multiple hash functions.
        """
        for i in range(self.hash_count):
            hash_result = int(hashlib.md5((item + str(i)).encode()).hexdigest(), 16)
            yield hash_result % self.size

    def add(self, item: str) -> None:
        """
        Add an item to the Bloom filter.

        :param item: The item to add.
        """
        for hash_value in self._hashes(item):
            self.bit_array[hash_value] = 1

    def check(self, item: str) -> bool:
        """
        Check if an item is in the Bloom filter.

        :param item: The item to check.
        :return: True if the item is possibly in the filter, False if it is definitely not.
        """
        return all(self.bit_array[hash_value] for hash_value in self._hashes(item))

    @staticmethod
    def optimal_parameters(n: int, p: float) -> tuple[int, int]:
        """
        Calculate optimal size and hash count for a given number of elements (n) and false positive
        probability (p).

        :param n: Number of elements expected to be added to the filter.
        :param p: Desired false positive probability.
        :return: A tuple containing the optimal size of the bit array and the number of hash functions.
        """
        m = -(n * math.log(p)) / (math.log(2) ** 2)
        k = (m / n) * math.log(2)
        return int(m), int(k)



# SPDX-FileCopyrightText: 2026 Greg Brandt <brandt.greg@gmail.com>
#
# SPDX-License-Identifier: Apache-2.0

import argparse
import sys
import subprocess
import tempfile
import logging
import requests
import time

logger = logging.getLogger(__name__)

DEFAULT_CHROME_DEBUG_PORT = 9222


class ChromeTool:
    def __init__(self) -> None:
        self.proc: subprocess.Popen | None = None

    def name(self) -> str:
        return "chrome"

    def start_chrome(
        self,
        headless: bool = False,
        chrome_debug_port: int = DEFAULT_CHROME_DEBUG_PORT,
        proxy_server: str | None = None,
        wait: bool = False,
    ) -> None:
        with tempfile.TemporaryDirectory() as tmpdir:
            # Build the command to start Chrome
            chrome_command = [
                "google-chrome",
                f"--remote-debugging-port={chrome_debug_port}",
                f"--user-data-dir={tmpdir}",
                "--no-first-run",
                "--no-default-browser-check",
            ]
            if headless:
                chrome_command.append("--headless")
            if proxy_server:
                chrome_command.append(f"--proxy-server={proxy_server}")
            # Start Chrome with the specified command
            logger.info("Starting Chrome with command: %s", " ".join(chrome_command))
            self.proc = subprocess.Popen(
                chrome_command,
                stdout=sys.stdout,
                stderr=sys.stderr,
            )
            if wait:
                self.proc.wait()

    def wait_for_chrome(
        self,
        chrome_debug_port: int = DEFAULT_CHROME_DEBUG_PORT,
        timeout: int | None = None,
    ) -> None:
        start_time = time.time()
        while timeout is None or (time.time() - start_time) < timeout:
            try:
                res = requests.get(f"http://localhost:{chrome_debug_port}/json")
                if res.status_code == 200:
                    logger.info("Chrome is ready on port %d", chrome_debug_port)
                    return
            except requests.ConnectionError:
                pass
            time.sleep(1)

    def stop_chrome(self) -> None:
        if self.proc:
            logger.info("Stopping Chrome with PID %d", self.proc.pid)
            self.proc.terminate()
            self.proc.wait()
            self.proc = None

    def configure(self, subparsers: argparse._SubParsersAction) -> None:
        parser = subparsers.add_parser(self.name(), help="Runs Chrome")
        parser.add_argument(
            "--headless",
            action="store_true",
            help="Run Chrome in headless mode",
        )
        parser.add_argument(
            "--chrome_debug_port",
            type=int,
            default=DEFAULT_CHROME_DEBUG_PORT,
            help="Open Chrome DevTools Protocol on this port",
        )
        parser.add_argument(
            "--proxy_server",
            help="Proxy server to use (e.g., 'http://localhost:8080')",
        )

    def run(self, args: argparse.Namespace) -> None:
        try:
            self.start_chrome(
                headless=args.headless,
                chrome_debug_port=args.chrome_debug_port,
                proxy_server=args.proxy_server,
                wait=True,
            )
        except KeyboardInterrupt:
            self.stop_chrome()



# SPDX-FileCopyrightText: 2026 Greg Brandt <brandt.greg@gmail.com>
#
# SPDX-License-Identifier: Apache-2.0

import argparse
from typing import Protocol


class Tool(Protocol):
    def name(self) -> str: ...
    def configure(self, subparsers: argparse._SubParsersAction) -> None: ...
    def run(self, args: argparse.Namespace) -> None: ...



# SPDX-FileCopyrightText: 2026 Greg Brandt <brandt.greg@gmail.com>
#
# SPDX-License-Identifier: Apache-2.0

from dataclasses import dataclass
from urllib.parse import urlparse
from avrokit import (
    URL,
    avro_reader,
    avro_schema,
    avro_writer,
    create_url_mapping,
    parse_url,
)
from datetime import datetime
from enum import Enum
from seleniumbase import SB
from typing import Generator, List
import argparse
import logging
import re
import time
import json
from rubbernecker.crawl.bloomfilter import BloomFilter
from .actions import (
    CrawlActionPlan,
    parse_crawl_action_plans,
    crawl_action,
    CrawlActionName,
)

logger = logging.getLogger(__name__)

SCHEMA = avro_schema(
    {
        "name": "Page",
        "type": "record",
        "fields": [
            {"name": "url", "type": "string"},
            {"name": "timestamp", "type": "long"},
            {"name": "body", "type": ["null", "string"], "default": None},
            {"name": "error", "type": ["null", "string"], "default": None},
            # TODO Support passing metadata from the request
            {
                "name": "metadata",
                "type": [
                    "null",
                    {
                        "type": "map",
                        "values": ["null", "string"],
                    },
                ],
                "default": None,
            },
        ],
    }
)


class InputFormat(Enum):
    TEXT = "text"
    JSON = "json"
    AVRO = "avro"


DEFAULT_INPUT_FORMAT: InputFormat = InputFormat.TEXT
DEFAULT_SLEEP_SUCCESS: int = 3
DEFAULT_SLEEP_ERROR: int = 10
DEFAULT_MAX_DEPTH: int = 0
DEFAULT_MAX_RETRIES: int = 0
DEFAULT_REPORT_INTERVAL: int = 100


@dataclass
class CrawlToolStats:
    count_input: int = 0
    count_output: int = 0
    count_error: int = 0


class CrawlTool:
    def name(self) -> str:
        return "crawl"

    def bloom_filter_key(self, url: str) -> str:
        """
        Generate a key for the Bloom filter based on the URL.

        :param url: The URL to generate a key for.
        :return: The generated key.
        """
        parsed_url = urlparse(url.lower())
        return f"{parsed_url.netloc}:{parsed_url.path}:{parsed_url.query}"

    def load_bloom_filter(self, output_url: URL) -> BloomFilter | None:
        """
        Load a Bloom filter from the output URL.

        :param output_url: URL to the output file.
        :return: A Bloom filter containing the URLs from the output file.
        """
        if output_url.exists():
            count = 0
            bloom_filter = BloomFilter()
            for url in output_url.expand():
                with avro_reader(url.with_mode("rb")) as reader:
                    for record in reader:
                        if (
                            isinstance(record, dict)
                            and "url" in record
                            and record.get("error") is None
                        ):
                            bloom_filter.add(self.bloom_filter_key(record["url"]))
                            count += 1
            if count > 0:
                logger.debug(
                    "Loaded %d URLs into Bloom filter from %s", count, output_url
                )
                return bloom_filter
        return None

    def load_requests(
        self,
        input_url: URL,
        input_format: InputFormat,
        bloom_filter: BloomFilter | None = None,
    ) -> Generator[str, None, None]:
        """
        Load requests from the input URL based on the specified format.

        :param input_url: URL to the input file.
        :param input_format: Format of the input file (text, JSON, or Avro).
        :param bloom_filter: Optional Bloom filter to check for already performed requests.
        :return: A generator yielding URLs from the input file.
        """
        logger.debug(
            "Loading requests from %s with format %s (bloom_filter=%s)",
            input_url,
            input_format,
            bloom_filter,
        )
        if input_format == InputFormat.TEXT:
            # Input is a text file with one URL per line
            with input_url.with_mode("r") as file:
                for line in file:
                    url = line.strip()
                    if not bloom_filter or not bloom_filter.check(
                        self.bloom_filter_key(url)
                    ):
                        yield url
        elif input_format == InputFormat.JSON:
            # Input is a JSONL file with one URL per line
            with input_url.with_mode("r") as file:
                for line in file:
                    data = json.loads(line)
                    url = data["url"]
                    if not bloom_filter or not bloom_filter.check(
                        self.bloom_filter_key(url)
                    ):
                        yield url
        elif input_format == InputFormat.AVRO:
            # Input is an Avro file
            with avro_reader(input_url.with_mode("rb")) as reader:
                for record in reader:
                    if isinstance(record, dict) and "url" in record:
                        url = record["url"]
                        if not bloom_filter or not bloom_filter.check(
                            self.bloom_filter_key(url)
                        ):
                            yield url

    def crawl(
        self,
        base_input_url: URL,
        base_output_url: URL,
        headless: bool = False,
        interactive: bool = False,
        input_format: InputFormat = DEFAULT_INPUT_FORMAT,
        sleep_success: int = DEFAULT_SLEEP_SUCCESS,
        sleep_error: int = DEFAULT_SLEEP_ERROR,
        load_actions: List[CrawlActionPlan] | None = None,
        crawl_actions: CrawlActionPlan | None = None,
        max_depth: int = DEFAULT_MAX_DEPTH,
        max_retries: int = DEFAULT_MAX_RETRIES,
        report_interval: int = DEFAULT_REPORT_INTERVAL,
        use_bloom_filter: bool = False,
        max_errors: int | None = None,
        proxy_server: str | None = None,
    ) -> CrawlToolStats:
        # Load the Bloom filter if needed
        bloom_filter: BloomFilter | None = None
        if use_bloom_filter:
            bloom_filter = self.load_bloom_filter(base_output_url)
            logger.debug("Loaded Bloom filter: %s", bloom_filter)

        stats = CrawlToolStats()
        first_request = True

        with SB(
            uc=True,
            test=True,
            incognito=True,
            headless=headless,
            proxy=proxy_server,
            locale="en-US",
        ) as sb:
            # Iterate over the input and output URLs
            for input_url, output_url in create_url_mapping(
                base_input_url, base_output_url
            ):
                logger.info("Mapping %s to %s", input_url, output_url)
                with avro_writer(output_url.with_mode("a+b"), SCHEMA) as writer:
                    # Execute requests
                    for url in self.load_requests(
                        input_url, input_format, bloom_filter=bloom_filter
                    ):
                        # Report progress
                        if (
                            stats.count_input > 0
                            and stats.count_input % report_interval == 0
                        ):
                            logger.info("%s", stats)

                        # Mark the timestamp and depth
                        stats.count_input += 1
                        timestamp = int(datetime.now().timestamp())  # seconds
                        depth = 0
                        retries = 0
                        crawling = False

                        while True:
                            try:
                                logger.info(
                                    "Crawling URL: %s (depth=%d, retries=%s)",
                                    url,
                                    depth,
                                    retries,
                                )

                                # Navigate to the URL
                                if not crawling:
                                    if first_request:
                                        sb.activate_cdp_mode(url)
                                        first_request = False
                                    elif sb.driver is None:
                                        raise Exception(
                                            "SeleniumBase driver is not initialized"
                                        )
                                    else:
                                        sb.driver.execute_cdp_cmd(
                                            "Page.navigate", {"url": url}
                                        )

                                # Perform load actions if provided
                                if load_actions:
                                    for plan in load_actions:
                                        if plan.should_run(url):
                                            plan_result = plan.run(sb.driver)
                                            if not plan_result:
                                                logger.error(
                                                    "Load actions failed for URL: %s",
                                                    url,
                                                )
                                                break

                                # Wait for the page to load
                                if interactive:
                                    # Pause if in interactive mode
                                    user_input = input(
                                        "Press any key to continue or 'q' to quit: "
                                    )
                                    if user_input.lower() == "q":
                                        logger.info("Exiting interactive mode.")
                                        return stats
                                else:
                                    # Sleep between requests
                                    time.sleep(sleep_success)

                                # Save the crawled data to the output URL
                                writer.append(
                                    {
                                        "url": sb.get_current_url(),
                                        "timestamp": timestamp,
                                        "body": sb.get_page_source(),
                                    }
                                )
                                stats.count_output += 1

                                # Check the current depth
                                if depth >= max_depth:
                                    logger.debug(
                                        "Maximum depth reached for URL: %s", url
                                    )
                                    break
                                depth += 1

                                # Run the crawl script to see if we should continue
                                # N.b. the action will be run and we'll expect the page to navigate to
                                # whatever is next. If the action returns False, we should stop the crawl
                                # and continue to the next URL. We should also limit the maximum depth of
                                # the crawl.
                                if crawl_actions and crawl_actions.should_run(url):
                                    # Set the crawling flag to True so we don't reload the page
                                    crawling = True
                                    if not crawl_actions.run(sb.driver):
                                        logger.debug("Crawl finished for URL: %s", url)
                                        break
                                else:
                                    # No crawl actions provided, just break
                                    break
                            except Exception as e:
                                # Sleep on error if not in interactive mode
                                if not interactive:
                                    time.sleep(sleep_error)

                                # Log the error and save it to the output URL
                                logger.error("Error crawling URL %s: %s", url, e)
                                writer.append(
                                    {
                                        "url": url,
                                        "timestamp": timestamp,
                                        "error": str(e),
                                    }
                                )
                                stats.count_error += 1

                                # Check if we should exit due to errors
                                if max_errors and stats.count_error >= max_errors:
                                    raise Exception(f"Max errors reached: {max_errors}")

                                # Check if we should retry
                                if retries >= max_retries:
                                    logger.debug(
                                        "Maximum retries reached for URL: %s", url
                                    )
                                    break
        return stats

    def configure(self, subparsers: argparse._SubParsersAction) -> None:
        parser = subparsers.add_parser(self.name(), help="Crawl web pages")
        parser.add_argument(
            "input_url",
            help="URL containing requests to crawl",
        )
        parser.add_argument(
            "output_url",
            help="URL to save the crawled data",
        )
        parser.add_argument(
            "--input_format",
            type=InputFormat,
            choices=list(InputFormat),
            default=DEFAULT_INPUT_FORMAT,
            help="Format of the input URL",
        )
        parser.add_argument(
            "--headless",
            action="store_true",
            help="Run Chrome in headless mode (default: supervised mode with visible browser)",
        )
        parser.add_argument(
            "--interactive",
            action="store_true",
            help="Run in interactive mode",
        )
        parser.add_argument(
            "--sleep_success",
            type=int,
            default=DEFAULT_SLEEP_SUCCESS,
            help="Sleep time after a successful request",
        )
        parser.add_argument(
            "--sleep_error",
            type=int,
            default=DEFAULT_SLEEP_ERROR,
            help="Sleep time after an error",
        )
        parser.add_argument(
            "--load_actions",
            help="Crawl action script to perform after page load",
        )
        parser.add_argument(
            "--sleep_after_load",
            type=float,
            help="Sleep for N seconds after page load (shortcut for creating a load action plan)",
        )
        parser.add_argument(
            "--crawl_actions",
            help="Crawl action script to perform during the crawl",
        )
        parser.add_argument(
            "--max_depth",
            type=int,
            default=DEFAULT_MAX_DEPTH,
            help="Maximum depth to crawl",
        )
        parser.add_argument(
            "--max_retries",
            type=int,
            default=DEFAULT_MAX_RETRIES,
            help="Maximum number of retries on error",
        )
        parser.add_argument(
            "--report_interval",
            type=int,
            default=DEFAULT_REPORT_INTERVAL,
            help="Report progress every N requests",
        )
        parser.add_argument(
            "--use_bloom_filter",
            action="store_true",
            help="Use a Bloom filter on output file to avoid duplicate requests",
        )
        parser.add_argument(
            "--max_errors",
            type=int,
            default=None,
            help="Maximum number of errors before exiting",
        )
        parser.add_argument(
            "--proxy_server",
            help="Proxy server URL (e.g., http://localhost:3128)",
        )

    def parse_load_actions(
        self,
        actionsfile: URL | None,
        sleep_after_load: float | None = None,
    ) -> List[CrawlActionPlan] | None:
        plans: List[CrawlActionPlan] = []
        if sleep_after_load is not None:
            plans.append(
                CrawlActionPlan(
                    url_pattern=re.compile(".*"),
                    actions=[
                        (crawl_action(CrawlActionName.SLEEP), [str(sleep_after_load)])
                    ],
                )
            )
        if actionsfile:
            with actionsfile.with_mode("r") as file:
                script = file.read()
                plans.extend(parse_crawl_action_plans(script))
        return plans if plans else None

    def parse_crawl_actions(self, actionsfile: URL | None) -> CrawlActionPlan | None:
        if actionsfile:
            with actionsfile.with_mode("r") as file:
                script = file.read()
                plans = parse_crawl_action_plans(script)
                if len(plans) != 1:
                    raise ValueError(
                        "Crawl action script must contain exactly one plan"
                    )
                return plans[0]
        return None

    def run(self, args: argparse.Namespace) -> None:
        load_actions_url = parse_url(args.load_actions) if args.load_actions else None
        crawl_actions_url = (
            parse_url(args.crawl_actions) if args.crawl_actions else None
        )
        stats = self.crawl(
            base_input_url=parse_url(args.input_url),
            base_output_url=parse_url(args.output_url),
            headless=args.headless,
            interactive=args.interactive,
            input_format=args.input_format,
            sleep_success=args.sleep_success,
            sleep_error=args.sleep_error,
            load_actions=self.parse_load_actions(
                load_actions_url,
                sleep_after_load=args.sleep_after_load,
            ),
            crawl_actions=self.parse_crawl_actions(crawl_actions_url),
            max_depth=args.max_depth,
            max_retries=args.max_retries,
            report_interval=args.report_interval,
            use_bloom_filter=args.use_bloom_filter,
            max_errors=args.max_errors,
            proxy_server=args.proxy_server,
        )
        logger.info("%s (done)", stats)



# SPDX-FileCopyrightText: 2026 Greg Brandt <brandt.greg@gmail.com>
#
# SPDX-License-Identifier: Apache-2.0

from .chrome import ChromeTool
from .proxy import ProxyTool

__all__ = ["ChromeTool", "ProxyTool"]



# SPDX-FileCopyrightText: 2026 Greg Brandt <brandt.greg@gmail.com>
#
# SPDX-License-Identifier: Apache-2.0

from dataclasses import dataclass
from enum import Enum
from typing import Any, Dict, List, Protocol, Tuple
import logging
import re
import time

logger = logging.getLogger(__name__)


class CrawlActionName(Enum):
    SLEEP = "SLEEP"
    INPUT = "INPUT"
    SCROLL = "SCROLL"
    CLICK = "CLICK"
    CLICK_IF_EXISTS = "CLICK_IF_EXISTS"


# Type alias for SeleniumBase Driver
CrawlDriver = Any


class CrawlAction(Protocol):
    def name(self) -> CrawlActionName: ...
    def run(self, driver: CrawlDriver, args: List[str]) -> bool: ...


class SleepCrawlAction:
    def name(self) -> CrawlActionName:
        return CrawlActionName.SLEEP

    def run(self, driver: CrawlDriver, args: List[str]) -> bool:
        if len(args) != 1:
            raise ValueError(
                "Sleep action requires exactly one argument: duration (seconds)"
            )
        try:
            duration = float(args[0])
            time.sleep(duration)
            return True
        except Exception as e:
            print(f"Error sleeping: {e}")
            return False


class InputCrawlAction:
    def name(self) -> CrawlActionName:
        return CrawlActionName.INPUT

    def run(self, driver: CrawlDriver, args: List[str]) -> bool:
        if len(args) < 2:
            raise ValueError(
                "Input action requires at least two arguments: selector value [value...]"
            )
        try:
            selector = args[0]
            value = " ".join(args[1:])
            driver.type(selector, value)
            return True
        except Exception as e:
            print(f"Error filling input: {e}")
            return False


class ScrollCrawlAction:
    def name(self) -> CrawlActionName:
        return CrawlActionName.SCROLL

    def run(self, driver: CrawlDriver, args: List[str]) -> bool:
        if len(args) != 1:
            raise ValueError("Scroll action requires exactly one argument: amount")
        try:
            amount = int(args[0])
            driver.execute_script(f"window.scrollBy(0, {amount})")
            return True
        except Exception as e:
            print(f"Error scrolling: {e}")
            return False


class ClickCrawlAction:
    def name(self) -> CrawlActionName:
        return CrawlActionName.CLICK

    def run(self, driver: CrawlDriver, args: List[str]) -> bool:
        if len(args) == 0:
            raise ValueError("Click action requires selector")
        try:
            selector = " ".join(args)
            logger.debug(f"Clicking on selector: {selector}")
            driver.click(selector)
            return True
        except Exception as e:
            logger.error(f"Error clicking: {e}")
            return False


class ClickIfExistsCrawlAction:
    def name(self) -> CrawlActionName:
        return CrawlActionName.CLICK_IF_EXISTS

    def run(self, driver: CrawlDriver, args: List[str]) -> bool:
        if len(args) == 0:
            raise ValueError("Click action requires selector")
        try:
            selector = " ".join(args)
            if not driver.is_element_present(selector):
                logger.debug(f"Selector not found: {selector}")
                return True
            logger.debug(f"Clicking on selector: {selector}")
            driver.click(selector)
            return True
        except Exception as e:
            logger.error(f"Error clicking: {e}")
            return False


ACTIONS: List[CrawlAction] = [
    SleepCrawlAction(),
    InputCrawlAction(),
    ScrollCrawlAction(),
    ClickCrawlAction(),
    ClickIfExistsCrawlAction(),
]

ACTION_NAMES: Dict[CrawlActionName, CrawlAction] = {
    action.name(): action for action in ACTIONS
}


def crawl_action(action_name: CrawlActionName) -> CrawlAction:
    """
    Get the crawl action by name.
    """
    action = ACTION_NAMES.get(action_name)
    if action is None:
        raise ValueError(f"Unknown action: {action_name}")
    return action


@dataclass
class CrawlActionPlan:
    url_pattern: re.Pattern
    actions: List[Tuple[CrawlAction, List[str]]]

    def should_run(self, url: str) -> bool:
        """
        Check if the action plan should run for the given URL.
        """
        return self.url_pattern.search(url) is not None

    def run(self, driver: CrawlDriver) -> bool:
        """
        Run the actions on the given driver.

        :param driver: The SeleniumBase driver to run the actions on.
        :return: True if the actions were run successfully, False otherwise.
        """
        for action, args in self.actions:
            if not action.run(driver, args):
                return False
        return True


def parse_crawl_action_plans(script: str) -> List[CrawlActionPlan]:
    """
    Parse a string of crawl actions into a dictionary of action names and their arguments.
    """
    acc: List[CrawlActionPlan] = []
    cur: CrawlActionPlan | None = None
    for line in script.strip().splitlines():
        line = line.strip()
        if line.startswith("[") and line.endswith("]"):
            if cur is not None:
                acc.append(cur)
            cur = CrawlActionPlan(
                url_pattern=re.compile(line[1:-1]),
                actions=[],
            )
        elif cur is not None:
            parts = line.split()
            action_name = CrawlActionName(parts[0].upper())
            action_args = parts[1:]
            action = ACTION_NAMES.get(action_name)
            if action is None:
                raise ValueError(f"Unknown action: {action_name}")
            cur.actions.append((action, action_args))
    if cur is not None:
        acc.append(cur)
    return acc



# SPDX-FileCopyrightText: 2026 Greg Brandt <brandt.greg@gmail.com>
#
# SPDX-License-Identifier: Apache-2.0

from typing import Generator, cast, Type
from avro.schema import Schema
import sys
import importlib
import pkgutil
import inspect
from abc import ABC, abstractmethod


class Parser(ABC):
    @abstractmethod
    def parse(self, record: object) -> Generator[object | None, None, None]: ...

    @abstractmethod
    def schema(self) -> Schema: ...


def list_parsers() -> list[Type[Parser]]:
    if not __package__:
        raise ValueError("Module name is not set. Cannot list parsers.")
    # Load all the modules to ensure that parser classes are registered
    module = importlib.import_module(__package__)
    if isinstance(module, str):
        module = importlib.import_module(module)
    results = {}
    for _, name, _ in pkgutil.walk_packages(module.__path__, module.__name__ + "."):
        try:
            results[name] = importlib.import_module(name)
        except Exception:
            pass  # Ignore modules that can't be imported
    # Find all classes that implement the Parser protocol
    classes: list[Type[Parser]] = []
    for mod in sys.modules.values():
        if mod and getattr(mod, "__name__", "").startswith(module.__name__):
            for _, obj in inspect.getmembers(mod, inspect.isclass):
                if (
                    isinstance(obj, type)
                    and obj.__module__.startswith(module.__name__)
                    and issubclass(obj, Parser)
                    and obj is not Parser
                ):
                    classes.append(cast(Type[Parser], obj))
    return classes



# SPDX-FileCopyrightText: 2026 Greg Brandt <brandt.greg@gmail.com>
#
# SPDX-License-Identifier: Apache-2.0

from typing import Any, Dict, Generator
from avrokit import avro_schema
from avro.schema import Schema
from .base import Parser
from bs4 import BeautifulSoup, Tag
from urllib.parse import urlparse
import logging

logger = logging.getLogger(__name__)


class StandardPageParser(Parser):
    def schema(self) -> Schema:
        return avro_schema(
            {
                "name": "StandardPage",
                "type": "record",
                "fields": [
                    {"name": "url", "type": "string"},
                    {"name": "timestamp", "type": "long"},
                    {"name": "content_length", "type": "int"},
                    {"name": "title", "type": ["null", "string"], "default": None},
                    {"name": "body_text", "type": ["null", "string"], "default": None},
                    {
                        "name": "headers",
                        "type": [
                            "null",
                            {
                                "type": "array",
                                "items": {
                                    "type": "record",
                                    "name": "Header",
                                    "fields": [
                                        {"name": "level", "type": "int"},
                                        {
                                            "name": "text",
                                            "type": ["null", "string"],
                                            "default": None,
                                        },
                                    ],
                                },
                            },
                        ],
                        "default": None,
                    },
                    {
                        "name": "links",
                        "type": [
                            "null",
                            {
                                "type": "array",
                                "items": {
                                    "type": "record",
                                    "name": "Link",
                                    "fields": [
                                        {
                                            "name": "text",
                                            "type": ["null", "string"],
                                            "default": None,
                                        },
                                        {
                                            "name": "url",
                                            "type": ["null", "string"],
                                            "default": None,
                                        },
                                        {
                                            "name": "external",
                                            "type": "boolean",
                                            "default": False,
                                        },
                                    ],
                                },
                            },
                        ],
                    },
                ],
            }
        )

    def _parse_headers(self, soup: BeautifulSoup) -> list[Dict[str, Any]]:
        headers: list[Dict[str, Any]] = []
        for header in soup.find_all(["h1", "h2", "h3", "h4", "h5", "h6"]):
            if isinstance(header, Tag):
                level = int(header.name[1])
                headers.append({"level": level, "text": header.get_text()})
        return headers

    def _parse_links(self, url: str, soup: BeautifulSoup) -> list[Dict[str, Any]]:
        parsed_url = urlparse(url)
        links: list[Dict[str, Any]] = []
        for link in soup.find_all("a"):
            if isinstance(link, Tag):
                href = link.get("href")
                if href:
                    parsed_href = urlparse(str(href))
                    links.append(
                        {
                            "text": link.get_text(),
                            "url": str(href),
                            "external": bool(parsed_href.netloc)
                            and (parsed_href.netloc != parsed_url.netloc),
                        }
                    )
        return links

    def parse(self, record: object) -> Generator[object | None, None, None]:
        if not isinstance(record, dict):
            logger.error("Record is not a dictionary: %s", record)
            return
        soup = BeautifulSoup(record["body"], "html.parser")
        body_text = soup.body.get_text() if soup.body else None
        yield {
            "url": record["url"],
            "timestamp": record["timestamp"],
            "title": soup.title.get_text() if soup.title else None,
            "content_length": len(record["body"]) if record["body"] else 0,
            "body_text": body_text,
            "headers": self._parse_headers(soup),
            "links": self._parse_links(record["url"], soup),
        }



#!/usr/bin/env python3

# SPDX-FileCopyrightText: 2026 Greg Brandt <brandt.greg@gmail.com>
#
# SPDX-License-Identifier: Apache-2.0

import logging
import argparse
from .base import Tool
from .parse import ParseTool
from .crawl import CrawlTool
from .browser import ProxyTool

TOOLS: list[Tool] = [
    CrawlTool(),
    ParseTool(),
    ProxyTool(),
]


def select_tool(tool_name: str):
    for tool in TOOLS:
        if tool.name() == tool_name:
            return tool
    raise ValueError(f"Tool {tool_name} not found.")


def configure_tools(subparsers: argparse._SubParsersAction) -> None:
    for tool in TOOLS:
        tool.configure(subparsers)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--debug", action="store_true")
    subparsers = parser.add_subparsers(dest="tool", required=True)
    configure_tools(subparsers)
    args = parser.parse_args()
    logging.basicConfig(
        format="%(levelname)s:%(message)s",
        level=logging.DEBUG if args.debug else logging.INFO,
    )
    try:
        tool = select_tool(args.tool)
        tool.run(args)
    except argparse.ArgumentError as e:
        parser.error(e.message)


if __name__ == "__main__":
    main()
